@@ -1,6 +1,7 @@
 import logging
 from datetime import datetime, date, time, timedelta, tzinfo
-from decimal import Decimal
+from decimal import Decimal, localcontext
+from dateutil.tz import tzoffset, tzutc
 import uuid
 from tds import *
 from tds import _Column
@@ -15,43 +16,25 @@ logger = logging.getLogger(__name__)
 # */
 def tds_set_column_type(tds, curcol, type):
     # set type
-    curcol.on_server.column_type = type
-    curcol.funcs = tds_get_column_funcs(tds, type)
-    curcol.column_type = tds_get_cardinal_type(type, curcol.column_usertype)
+    curcol.column_type = type
+    curcol.funcs = _get_handler(tds, type)
 
     # set size
     curcol.column_varint_size = tds_get_varint_size(tds, type)
     if curcol.column_varint_size == 0:
-        curcol.on_server.column_size = curcol.column_size = tds_get_size_by_type(type)
+        curcol.column_size = curcol.column_size = tds_get_size_by_type(type)
 
-def tds_get_column_funcs(tds, type):
+def _get_handler(tds, type):
     if type in (SYBNUMERIC, SYBDECIMAL):
         return NumericHandler
     elif type == SYBVARIANT:
-        if IS_TDS7_PLUS(tds):
-            return VariantHandler
+        return VariantHandler
     elif type in (SYBMSDATE, SYBMSTIME, SYBMSDATETIME2, SYBMSDATETIMEOFFSET):
         return MsDatetimeHandler
-    return DefaultHandler
-
-
-def tds_get_cardinal_type(datatype, usertype):
-    if datatype == XSYBVARBINARY:
-        return SYBVARBINARY
-    if datatype == XSYBBINARY:
-        return SYBBINARY
-    if datatype == SYBNTEXT:
-        return SYBTEXT
-    if datatype in (XSYBNVARCHAR, XSYBVARCHAR):
-        return SYBVARCHAR
-    if datatype in (XSYBNCHAR, XSYBCHAR):
-        return SYBCHAR
-    if datatype == SYB5INT8:
-        return SYBINT8
-    if datatype == SYBLONGBINARY:
-        if usertype in (USER_UNICHAR_TYPE, USER_UNIVARCHAR_TYPE):
-            return SYBTEXT
-    return datatype
+    elif type in (SYBDATETIME, SYBDATETIMN, SYBDATETIME4):
+        return DatetimeHandler
+    else:
+        return DefaultHandler
 
 def make_param(tds, name, value):
     column = _Column()
@@ -62,87 +45,45 @@ def make_param(tds, name, value):
         value = value.value
     if value is default:
         column.flags = fDefaultValue
-        col_type = XSYBVARCHAR
-        size = 1
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
         value = None
-    elif value is None:
-        col_type = XSYBVARCHAR
-        size = 1
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
-    elif isinstance(value, int):
-        if -2**31 <= value <= 2**31 -1:
-            col_type = SYBINTN
-            size = 4
+    if value is None:
+        handler = DefaultHandler
+    elif isinstance(value, (int, long)):
+        if -2**63 <= value <= 2*63-1:
+            handler = DefaultHandler
+        elif -10**38+1 <= value <= 10**38-1:
+            value = Decimal(value)
+            handler = NumericHandler
         else:
-            col_type = SYBINT8
-            size = 8
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
+            raise DataError('Numeric value out or range')
     elif isinstance(value, float):
-        col_type = SYBFLTN
-        size = 8
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
+        handler = DefaultHandler
     elif isinstance(value, Binary):
-        if len(value) > 8000:
-            if IS_TDS72_PLUS(tds):
-                col_type = XSYBVARBINARY
-                column.column_varint_size = 8 # nvarchar(max)
-            else:
-                col_type = SYBIMAGE
-                column.column_varint_size = tds_get_varint_size(tds, col_type)
-        else:
-            col_type = XSYBVARBINARY
-            column.column_varint_size = tds_get_varint_size(tds, col_type)
-        size = len(value)
+        handler = DefaultHandler
     elif isinstance(value, (str, unicode)):
-        if len(value) > 4000:
-            if IS_TDS72_PLUS(tds):
-                col_type = XSYBNVARCHAR
-                column.column_varint_size = 8 # nvarchar(max)
-            else:
-                col_type = SYBNTEXT
-                column.column_varint_size = tds_get_varint_size(tds, col_type)
-        else:
-            col_type = XSYBNVARCHAR
-            column.column_varint_size = tds_get_varint_size(tds, col_type)
-        size = len(value) * 2
-        column.char_codec = ucs2_codec
+        handler = DefaultHandler
     elif isinstance(value, datetime):
-        col_type = SYBDATETIMN
-        size = 8
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
+        if IS_TDS73_PLUS(tds):
+            handler = MsDatetimeHandler
+        else:
+            handler = DatetimeHandler
     elif isinstance(value, date):
         if IS_TDS73_PLUS(tds):
-            col_type = SYBMSDATE
+            handler = MsDatetimeHandler
         else:
-            col_type = SYBDATETIME
-        size = 1
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
+            handler = DatetimeHandler
     elif isinstance(value, time):
-        if IS_TDS73_PLUS(tds):
-            col_type = SYBMSTIME
-            column.precision = 7
-        else:
-            col_type = SYBDATETIME
-        size = 1
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
+        if not IS_TDS73_PLUS(tds):
+            raise DataError('Time type is not supported on MSSQL 2005 and lower')
+        handler = MsDatetimeHandler
     elif isinstance(value, Decimal):
-        col_type = SYBDECIMAL
-        _, digits, exp = value.as_tuple()
-        size = 12
-        column.column_scale = -exp
-        column.column_prec = max(len(digits), column.column_scale)
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
+        handler = NumericHandler
     elif isinstance(value, uuid.UUID):
-        col_type = SYBUNIQUE
-        size = 16
-        column.column_varint_size = tds_get_varint_size(tds, col_type)
+        handler = DefaultHandler
     else:
-        raise Exception('NotSupportedError: Unable to determine database type for value: {0}'.format(repr(value)))
-    column.on_server.column_type = col_type
-    column.column_size = column.on_server.column_size = size
-    column.value = value
-    column.funcs = tds_get_column_funcs(tds, col_type)
+        raise DataError('Parameter type is not supported: {0}'.format(repr(value)))
+    column.funcs = handler
+    column.value = column.funcs.from_python(tds, column, value)
     return column
 
 class DefaultHandler(object):
@@ -166,13 +107,13 @@ class DefaultHandler(object):
         elif vs == 0:
             col.column_size = tds_get_size_by_type(col.column_type)
 
-        if IS_TDS71_PLUS(tds) and is_collate_type(col.on_server.column_type):
+        if IS_TDS71_PLUS(tds) and is_collate_type(col.column_type):
             # based on true type as sent by server
             col.column_collation = r.get_collation()
             col.char_codec = col.column_collation.get_codec()
 
         # Only read table_name for blob columns (eg. not for SYBLONGBINARY)
-        if is_blob_type(col.on_server.column_type):
+        if is_blob_type(col.column_type):
             # discard this additional byte
             if IS_TDS72_PLUS(tds):
                 num_parts = r.get_byte()
@@ -181,7 +122,7 @@ class DefaultHandler(object):
                     col.table_name = r.read_ucs2(r.get_smallint())
             else:
                 col.table_name = r.read_ucs2(r.get_smallint())
-        elif IS_TDS72_PLUS(tds) and col.on_server.column_type == SYBMSXML:
+        elif IS_TDS72_PLUS(tds) and col.column_type == SYBMSXML:
             has_schema = r.get_byte()
             if has_schema:
                 # discard schema informations
@@ -219,8 +160,8 @@ class DefaultHandler(object):
                     val = decoder.decode(val)
                 chunk_handler.new_chunk(val)
 
-    @staticmethod
-    def get_data(tds, curcol):
+    @classmethod
+    def get_data(cls, tds, curcol):
         logger.debug("tds_get_data: type %d, varint size %d" % (curcol.column_type, curcol.column_varint_size))
         r = tds._reader
         cvs = curcol.column_varint_size
@@ -245,8 +186,7 @@ class DefaultHandler(object):
             if colsize == 0:
                 colsize = -1;
         elif cvs == 8:
-            curcol.value = DefaultHandler._tds72_get_varmax(tds, curcol)
-            return
+            return cls._tds72_get_varmax(tds, curcol)
         elif cvs == 2:
             colsize = r.get_smallint()
         elif cvs == 1:
@@ -258,14 +198,11 @@ class DefaultHandler(object):
             colsize = tds_get_size_by_type(curcol.column_type)
         else:
             colsize = -1;
-        if IS_TDSDEAD(tds):
-            raise Exception('TDS_FAIL')
 
         logger.debug("tds_get_data(): wire column size is %d" % colsize)
         # set NULL flag in the row buffer
         if colsize < 0:
-            curcol.value = None
-            return
+            return None
 
         #
         # We're now set to read the data from the wire.
@@ -280,24 +217,23 @@ class DefaultHandler(object):
             # TODO this can lead to a big waste of memory
             new_blob_size = colsize
             if new_blob_size == 0:
-                curcol.value = ''
-                return
+                return ''
 
             # read the data
             if curcol.char_codec:
-                curcol.value = tds_get_char_data(tds, colsize, curcol)
+                return tds_get_char_data(tds, colsize, curcol)
             else:
                 assert colsize == new_blob_size
-                curcol.value = r.readall(colsize)
+                return r.readall(colsize)
         else: # non-numeric and non-blob
             if curcol.char_codec:
                 if colsize == 0:
-                    curcol.value = u''
+                    value = u''
                 elif curcol.char_codec:
-                    curcol.value = curcol.char_codec.decode(r.readall(colsize))[0]
+                    value = curcol.char_codec.decode(r.readall(colsize))[0]
                 else:
-                    curcol.value = r.readall(colsize)
-                curcol.cur_size = len(curcol.value)
+                    value = r.readall(colsize)
+                curcol.cur_size = len(value)
             else:
                 #
                 # special case, some servers seem to return more data in some conditions 
@@ -307,24 +243,139 @@ class DefaultHandler(object):
                 if colsize > curcol.column_size:
                     discard_len = colsize - curcol.column_size
                     colsize = curcol.column_size
-                curcol.value = r.readall(colsize)
+                value = r.readall(colsize)
                 if discard_len > 0:
                     r.skip(discard_len)
 
             # pad (UNI)CHAR and BINARY types
-            fillchar = '\0'
-            if curcol.column_type in (SYBCHAR, XSYBCHAR) and (curcol.column_size == curcol.on_server.column_size):
-                # FIXME use client charset
-                fillchar = ' '
+            #fillchar = '\0'
+            #if curcol.column_type in (SYBCHAR, XSYBCHAR):
+            #    fillchar = ' '
             # extra handling for SYBLONGBINARY
-            if curcol.column_type == SYBLONGBINARY and curcol.column_usertype == USER_UNICHAR_TYPE or\
-                    curcol.column_type in (SYBCHAR, XSYBCHAR) and (curcol.column_size == curcol.on_server.column_size) or\
-                    curcol.column_type in (SYBBINARY, XSYBBINARY):
+            #if curcol.column_type == SYBLONGBINARY and curcol.column_usertype == USER_UNICHAR_TYPE or\
+            #        curcol.column_type in (SYBCHAR, XSYBCHAR) and (curcol.column_size == curcol.column_size) or\
+            #        curcol.column_type in (SYBBINARY, XSYBBINARY):
 
-                        if colsize < curcol.column_size:
-                            curcol.value.extend(fillchar*(curcol.column_size - colsize))
-                        colsize = curcol.column_size
-            curcol.value = to_python(tds, curcol.value, curcol.column_type, curcol.column_size)
+            #            if colsize < curcol.column_size:
+            #                value.extend(fillchar*(curcol.column_size - colsize))
+            #            colsize = curcol.column_size
+            return to_python(tds, value, curcol.column_type, curcol.column_size)
+
+    @staticmethod
+    def from_python(tds, col, value):
+        if value is None:
+            col.column_type = XSYBVARCHAR
+            col.column_size = 1
+            col.column_varint_size = tds_get_varint_size(tds, col.column_type)
+        elif isinstance(value, (int, long)):
+            col.column_type = SYBINTN
+            if -2**31 <= value <= 2**31 -1:
+                col.column_size = 4
+            elif -2**63 <= value <= 2*63-1:
+                col.column_size = 8
+            else:
+                raise DataError('Numeric value out or range')
+            col.column_varint_size = tds_get_varint_size(tds, col.column_type)
+        elif isinstance(value, float):
+            col.column_type = SYBFLTN
+            col.column_size = 8
+            col.column_varint_size = tds_get_varint_size(tds, col.column_type)
+        elif isinstance(value, Binary):
+            if len(value) > 8000:
+                if IS_TDS72_PLUS(tds):
+                    col.column_type = XSYBVARBINARY
+                    col.column_varint_size = 8 # nvarchar(max)
+                else:
+                    col.column_type = SYBIMAGE
+                    col.column_varint_size = tds_get_varint_size(tds, col.column_type)
+            else:
+                col.column_type = XSYBVARBINARY
+                col.column_varint_size = tds_get_varint_size(tds, col.column_type)
+            col.column_size = len(value)
+        elif isinstance(value, (str, unicode)):
+            if len(value) > 4000:
+                if IS_TDS72_PLUS(tds):
+                    col.column_type = XSYBNVARCHAR
+                    col.column_varint_size = 8 # nvarchar(max)
+                else:
+                    col.column_type = SYBNTEXT
+                    col.column_varint_size = tds_get_varint_size(tds, col.column_type)
+            else:
+                col.column_type = XSYBNVARCHAR
+                col.column_varint_size = tds_get_varint_size(tds, col.column_type)
+            col.column_size = len(value) * 2
+            col.char_codec = ucs2_codec
+        elif isinstance(value, uuid.UUID):
+            col.column_type = SYBUNIQUE
+            col.column_size = 16
+            col.column_varint_size = tds_get_varint_size(tds, col.column_type)
+        return value
+
+    @staticmethod
+    def get_declaration(tds, col):
+        max_len = 8000 if IS_TDS7_PLUS(tds) else 255
+        t = col.column_type
+        if t in (XSYBCHAR, SYBCHAR):
+            return "CHAR(%d)" % col.size
+        elif t in (SYBVARCHAR, XSYBVARCHAR):
+            if col.column_varint_size == 8:
+                return "VARCHAR(MAX)"
+            else:
+                return "VARCHAR(%d)" % col.column_size
+        elif t == SYBINT1:
+            return "TINYINT"
+        elif t == SYBINT2:
+            return "SMALLINT"
+        elif t == SYBINT4 or t == SYBINTN and col.column_size == 4:
+            return "INT"
+        elif t == SYBINT8 or t == SYBINTN and col.column_size == 8:
+            # TODO even for Sybase ??
+            return "BIGINT"
+        elif t == SYBFLT8 or t == SYBFLTN and col.column_size == 8:
+            return "FLOAT"
+        elif t == SYBBIT:
+            return "BIT"
+        elif t == SYBTEXT:
+            return "TEXT"
+        elif t == (SYBLONGBINARY, # TODO correct ??
+                SYBIMAGE):
+            return "IMAGE"
+        elif t == SYBMONEY4:
+            return "SMALLMONEY"
+        elif t == SYBMONEY:
+            return "MONEY"
+        elif t == SYBREAL:
+            return "REAL"
+        elif t in (SYBBINARY, XSYBBINARY):
+            return "BINARY(%d)" % min(size, max_len)
+        elif t in (SYBVARBINARY, XSYBVARBINARY):
+            if col.column_varint_size == 8:
+                return "VARBINARY(MAX)"
+            else:
+                return "VARBINARY(%u)" % min(col.column_size, max_len)
+        elif t == SYBUNIQUE:
+            if IS_TDS7_PLUS(tds):
+                return "UNIQUEIDENTIFIER"
+        elif t == SYBNTEXT:
+            if IS_TDS7_PLUS(tds):
+                return "NTEXT"
+        elif t in (SYBNVARCHAR, XSYBNVARCHAR):
+            if col.column_varint_size == 8:
+                return "NVARCHAR(MAX)"
+            elif IS_TDS7_PLUS(tds):
+                return "NVARCHAR(%u)" % max(col.column_size//2, 1)
+        elif t == XSYBNCHAR:
+            if IS_TDS7_PLUS(tds):
+                return "NCHAR(%u)" % min(col.column_size//2, 4000)
+        elif t == SYBVARIANT:
+            if IS_TDS7_PLUS(tds):
+                return "SQL_VARIANT"
+        # nullable types should not occur here...
+        elif t in (SYBMONEYN, SYBDATETIMN, SYBBITN):
+            assert False
+            # TODO...
+        else:
+            raise Exception("Unknown type %d", t)
 
     @staticmethod
     def put_info(tds, col):
@@ -343,8 +394,8 @@ class DefaultHandler(object):
             w.put_smallint(-1)
 
         # TDS7.1 output collate information
-        if IS_TDS71_PLUS(tds) and is_collate_type(col.on_server.column_type):
-            w.put_collation(tds.collation)
+        if IS_TDS71_PLUS(tds) and is_collate_type(col.column_type):
+            w.put_collation(tds.conn.collation)
 
     @staticmethod
     def put_data(tds, curcol):
@@ -403,10 +454,10 @@ class DefaultHandler(object):
                 w.put_byte(size)
             elif vs == 0:
                 # TODO should be column_size
-                colsize = tds_get_size_by_type(curcol.on_server.column_type)
+                colsize = tds_get_size_by_type(curcol.column_type)
 
             # put real data
-            column_type = curcol.on_server.column_type
+            column_type = curcol.column_type
             if column_type == SYBINTN and size == 4 or column_type == SYBINT4:
                 w.put_int(value)
             elif column_type == SYBINTN and size == 8 or column_type == SYBINT8:
@@ -415,10 +466,6 @@ class DefaultHandler(object):
                 w.write(value)
             elif column_type in (XSYBVARBINARY, XSYBBINARY):
                 w.write(value)
-            elif column_type in (SYBDATETIME, SYBDATETIMN):
-                days = (value - MsDatetimeHandler._base_date).days
-                tm = (value.hour * 60 * 60 + value.minute * 60 + value.second)*300 + value.microsecond/1000/3
-                w.write(TDS_DATETIME.pack(days, tm))
             elif column_type == SYBFLTN and size == 8 or column_type == SYBFLT8:
                 w.write(_SYBFLT8_STRUCT.pack(value))
             elif column_type == SYBNTEXT:
@@ -470,9 +517,6 @@ def to_python(tds, data, type, length):
         val = Decimal(val)/10000
         return val
 
-    elif type in (SYBDATETIME, SYBDATETIME4, SYBDATETIMN):
-        return tds_datecrack(type, data)
-
     elif type in (SYBVARCHAR, SYBCHAR, SYBTEXT, SYBBINARY,\
             SYBNVARCHAR, XSYBVARCHAR, XSYBNVARCHAR, XSYBCHAR, XSYBNCHAR,\
             XSYBVARBINARY, XSYBBINARY, SYBVARBINARY):
@@ -502,20 +546,21 @@ class NumericHandler(object):
     def ms_parse_numeric(positive, buf, scale):
         val = reduce(lambda acc, val: acc*256 + ord(val), reversed(buf), 0)
         val = Decimal(val)
-        if not positive:
-            val *= -1
-        val /= 10 ** scale
+        with localcontext() as ctx:
+            ctx.prec = 38
+            if not positive:
+                val *= -1
+            val /= 10 ** scale
         return val
 
-    @staticmethod
-    def get_data(tds,  curcol):
+    @classmethod
+    def get_data(cls, tds,  curcol):
         r = tds._reader
         colsize = r.get_byte()
 
         # set NULL flag in the row buffer
         if colsize <= 0:
-            curcol.value = None
-            return
+            return None
 
         #
         # Since these can be passed around independent
@@ -528,13 +573,13 @@ class NumericHandler(object):
 
         # server is going to crash freetds ??
         # TODO close connection it server try to do so ??
-        if colsize > NumericHandler.MAX_NUMERIC:
+        if colsize > cls.MAX_NUMERIC:
             raise Exception('TDS_FAIL')
         positive = r.get_byte()
         buf = r.readall(colsize - 1)
 
         if IS_TDS7_PLUS(tds):
-            curcol.value = NumericHandler.ms_parse_numeric(positive, buf, scale)
+            return cls.ms_parse_numeric(positive, buf, scale)
         else:
             raise Exception('not supported')
 
@@ -554,6 +599,30 @@ class NumericHandler(object):
         31, 31, 31, 32, 32, 33, 33, 33
         ]
 
+    @staticmethod
+    def from_python(tds, col, value):
+        if not (-10**38+1 <= value <= 10**38-1):
+            raise DataError('Decimal value is out of range')
+        value = value.normalize()
+        col.column_type = SYBDECIMAL
+        _, digits, exp = value.as_tuple()
+        if exp > 0:
+            col.column_scale = 0
+            col.column_prec = len(digits) + exp
+        else:
+            col.column_scale = -exp
+            col.column_prec = max(len(digits), col.column_scale)
+        if col.column_prec > 38:
+            raise DataError('Precision of decimal value is out of range')
+        return value
+
+    @staticmethod
+    def get_declaration(tds, col):
+        if col.column_type == SYBNUMERIC:
+            return "NUMERIC(%d,%d)" % (col.column_prec, col.column_scale)
+        elif col.column_type == SYBDECIMAL:
+            return "DECIMAL(%d,%d)" % (col.column_prec, col.column_scale)
+
     @staticmethod
     def put_info(tds, col):
         w = tds._writer
@@ -570,13 +639,15 @@ class NumericHandler(object):
         val = col.value
         positive = 1 if val > 0 else 0
         w.put_byte(positive) # sign
-        if not positive:
-            val *= -1
-        size -= 1
-        val = long(val * (10 ** scale))
+        with localcontext() as ctx:
+            ctx.prec = 38
+            if not positive:
+                val *= -1
+            size -= 1
+            val = long(val * (10 ** scale))
         for i in range(size):
             w.put_byte(val % 256)
-            val /= 256
+            val //= 256
         assert val == 0
 
 #
@@ -598,10 +669,9 @@ class VariantHandler(object):
 
         # NULL
         try:
-            curcol.value = None
             if colsize < 2:
                 r.skip(colsize)
-                return
+                return None
 
             type = r.get_byte();
             info_len = r.get_byte()
@@ -628,8 +698,7 @@ class VariantHandler(object):
                     raise Exception('TDS_FAIL')
                 positive = r.get_byte()
                 buf = r.readall(colsize - 1)
-                curcol.value = NumericHandler.ms_parse_numeric(positive, buf, scale)
-                return
+                return NumericHandler.ms_parse_numeric(positive, buf, scale)
             varint = 0 if type == SYBUNIQUE else tds_get_varint_size(tds, type)
             if varint != info_len:
                 raise Exception('TDS_FAIL')
@@ -648,7 +717,7 @@ class VariantHandler(object):
                 else:
                     data = r.readall(colsize)
             colsize = 0
-            curcol.value = to_python(tds, data, type, colsize)
+            return to_python(tds, data, type, colsize)
         except:
             r.skip(colsize)
             raise
@@ -723,7 +792,6 @@ XSYBVARCHAR	MS	2	0	1	1	0	0	0	1	-1	0
 # there are some type that allow size 0 or a constants (SYBDATEN, SYBUNIQUE)
 # some type (BITN, DATEN, UNIQUE, MSDATE) have size but are nullable
 # tds_get_conversion_type from nullable to not nullable
-# tds_get_cardinal_type ??
 
 # $Id: types.txt,v 1.5 2011/05/12 19:40:57 freddy77 Exp $
 '''
@@ -768,11 +836,11 @@ exec(gen_get_varint_size())
 # Get column size for wire
 #
 def tds_fix_column_size(tds, curcol):
-    size = curcol.on_server.column_size
+    size = curcol.column_size
 
     if not size:
         size = curcol.column_size
-        if is_unicode_type(curcol.on_server.column_type):
+        if is_unicode_type(curcol.column_type):
             size *= 2
 
     vs = curcol.column_varint_size
@@ -781,52 +849,31 @@ def tds_fix_column_size(tds, curcol):
     elif vs == 1:
         size = max(min(size, 255), 1)
     elif vs == 2:
-        if curcol.on_server.column_type in (XSYBNVARCHAR, XSYBNCHAR):
+        if curcol.column_type in (XSYBNVARCHAR, XSYBNCHAR):
             mn = 2
         else:
             mn = 1
         size = max(min(size, 8000), mn)
     elif vs == 4:
-        if curcol.on_server.column_type == SYBNTEXT:
+        if curcol.column_type == SYBNTEXT:
             size = max(min(size, 0x7ffffffe), 2)
         else:
             size = max(min(size, 0x7fffffff), 1)
-    #return curcol->on_server.column_size = size
     return size
 
 def tds_convert_string(tds, char_codec, s):
     return char_codec.encode(s)[0]
 
-ZERO = timedelta(0)
-
-# A class building tzinfo objects for fixed-offset time zones.
-# Note that FixedOffset(0, "UTC") is a different way to build a
-# UTC tzinfo object.
-
-class FixedOffset(tzinfo):
-    """Fixed offset in minutes east from UTC."""
-
-    def __init__(self, offset, name):
-        self.__offset = timedelta(minutes = offset)
-        self.__name = name
-
-    def utcoffset(self, dt):
-        return self.__offset
-
-    def tzname(self, dt):
-        return self.__name
-
-    def dst(self, dt):
-        return ZERO
+_utc = tzutc()
 
 class MsDatetimeHandler(object):
     @staticmethod
     def get_info(tds, col):
         r = tds._reader
-        col.column_scale = col.column_prec = 0
+        col.scale = col.prec = 0
         if col.column_type != SYBMSDATE:
-            col.column_scale = col.column_prec = r.get_byte()
-            if col.column_prec > 7:
+            col.scale = col.prec = r.get_byte()
+            if col.prec > 7:
                 raise Exception('TDS_FAIL')
 
     @staticmethod
@@ -834,8 +881,7 @@ class MsDatetimeHandler(object):
         r = tds._reader
         size = r.get_byte()
         if size == 0:
-            col.value = None
-            return
+            return None
 
         if col.column_type == SYBMSDATETIMEOFFSET:
             size -= 2
@@ -852,47 +898,73 @@ class MsDatetimeHandler(object):
                 raise Exception('TDS_FAIL')
             time_buf = r.readall(size)
             val = reduce(lambda acc, val: acc * 256 + ord(val), reversed(time_buf), 0)
-            for i in range(col.column_prec, 7):
-                val *= 10
+            val *= 10**(7-col.prec)
             nanoseconds = val*100
 
         # get date part
         days = 0
         if col.column_type != SYBMSTIME:
             date_buf = r.readall(3)
-            val = reduce(lambda acc, val: acc * 256 + ord(val), reversed(date_buf), 0)
-            days = val - 693595
+            days = reduce(lambda acc, val: acc * 256 + ord(val), reversed(date_buf), 0)
 
         # get time offset
-        tz = None
-        if col.column_type == SYBMSDATETIMEOFFSET:
+        if col.column_type == SYBMSTIME:
+            hours = nanoseconds//1000000000//60//60
+            nanoseconds -= hours*60*60*1000000000
+            minutes = nanoseconds/1000000000//60
+            nanoseconds -= minutes*60*1000000000
+            seconds = nanoseconds//1000000000
+            nanoseconds -= seconds*1000000000
+            return time(hours, minutes, seconds, nanoseconds//1000)
+        elif col.column_type == SYBMSDATE:
+            return date(1, 1, 1) + timedelta(days=days)
+        elif col.column_type == SYBMSDATETIME2:
+            return datetime(1, 1, 1) + timedelta(days=days, microseconds=nanoseconds//1000)
+        elif col.column_type == SYBMSDATETIMEOFFSET:
             offset = r.get_smallint()
             if offset > 840 or offset < -840:
                 raise Exception('TDS_FAIL')
-            tz = FixedOffset(offset, '')
+            tz = tzoffset('', offset*60)
+            return (datetime(1, 1, 1, tzinfo=_utc) + timedelta(days=days, microseconds=nanoseconds//1000)).astimezone(tz)
 
-        if col.column_type == SYBMSTIME:
-            hours = nanoseconds/1000000000/60/60
-            nanoseconds -= hours*60*60*1000000000
-            minutes = nanoseconds/1000000000/60
-            nanoseconds -= minutes*60*1000000000
-            seconds = nanoseconds/1000000000
-            nanoseconds -= seconds*1000000000
-            col.value = time(hours, minutes, seconds, nanoseconds/1000)
-        elif col.column_type == SYBMSDATE:
-            col.value = date(1900, 1, 1) + timedelta(days=days)
-        else:
-            col.value = datetime(1900, 1, 1, tzinfo=tz) + timedelta(days=days, microseconds=nanoseconds/1000)
+    @staticmethod
+    def from_python(tds, col, value):
+        if isinstance(value, datetime):
+            if value.tzinfo:
+                col.column_type = SYBMSDATETIMEOFFSET
+            else:
+                col.column_type = SYBMSDATETIME2
+            col.prec = col.scale = 6
+        elif isinstance(value, date):
+            col.column_type = SYBMSDATE
+        elif isinstance(value, time):
+            if value.tzinfo:
+                col.column_type = SYBMSDATETIMEOFFSET
+            else:
+                col.column_type = SYBMSTIME
+            col.prec = col.scale = 6
+        return value
+
+    @staticmethod
+    def get_declaration(tds, col):
+        t = col.column_type
+        if t == SYBMSTIME:
+            return "TIME({0})".format(col.scale)
+        elif t == SYBMSDATE:
+            return "DATE"
+        elif t == SYBMSDATETIME2:
+            return "DATETIME2({0})".format(col.scale)
+        elif t == SYBMSDATETIMEOFFSET:
+            return "DATETIMEOFFSET({0})".format(col.scale)
 
     @staticmethod
     def put_info(tds, col):
         w = tds._writer
-        # TODO precision
-        if col.on_server.column_type != SYBMSDATE:
-            w.put_byte(7)
+        if col.column_type != SYBMSDATE:
+            w.put_byte(col.prec)
 
-    _base_date = datetime(1900, 1, 1)
     _base_date2 = datetime(1, 1, 1)
+    _base_date2_utc = datetime(1, 1, 1, tzinfo=_utc)
 
     _precision_to_len = {
             0: 3,
@@ -912,25 +984,106 @@ class MsDatetimeHandler(object):
             w.put_byte(0)
             return
 
-        # TODO precision
         value = col.value
         parts = []
-        if col.on_server.column_type != SYBMSDATE:
+        tzinfo = None
+        if col.column_type != SYBMSDATE:
+            tzinfo = value.tzinfo
+            if tzinfo:
+                value = value.astimezone(_utc)
             t = value
             secs = t.hour*60*60 + t.minute*60 + t.second
-            val = (secs * 10**7 + t.microsecond*10)/(10**(7-col.precision))
-            parts.append(struct.pack('<Q', val)[:MsDatetimeHandler._precision_to_len[col.precision]])
-        if col.on_server.column_type != SYBMSTIME:
+            val = (secs * 10**7 + t.microsecond*10)//(10**(7-col.prec))
+            parts.append(struct.pack('<Q', val)[:MsDatetimeHandler._precision_to_len[col.prec]])
+        if col.column_type != SYBMSTIME:
             if type(value) == date:
                 value = datetime.combine(value, time(0,0,0))
-            days = (value - MsDatetimeHandler._base_date2).days
+            if tzinfo:
+                days = (value - MsDatetimeHandler._base_date2_utc).days
+            else:
+                days = (value - MsDatetimeHandler._base_date2).days
             buf = struct.pack('<l', days)[:3]
             parts.append(buf)
-        if col.on_server.column_type == SYBMSDATETIMEOFFSET:
-            parts.append(struct.pack('<H', value.utcoffset()))
+        if col.column_type == SYBMSDATETIMEOFFSET:
+            assert tzinfo
+            parts.append(struct.pack('<h', col.value.utcoffset().total_seconds()//60))
         size = reduce(lambda a, b: a + len(b), parts, 0)
-        parts.insert(0, chr(size))
-        w.write(b''.join(parts))
+        w.put_byte(size)
+        for part in parts:
+            w.write(part)
+
+class DatetimeHandler(object):
+    _base_date = datetime(1900, 1, 1)
+    _min_date = datetime(1753, 1, 1, 0, 0, 0)
+    _max_date = datetime(9999, 12, 31, 23, 59, 59, 997000)
+    @staticmethod
+    def get_info(tds, col):
+        r = tds._reader
+        if col.column_type == SYBDATETIMN:
+            col.size = r.get_byte()
+        elif col.column_type == SYBDATETIME:
+            col.size = 8
+        elif col.column_type == SYBDATETIME4:
+            col.size = 4
+        else:
+            assert False
+
+    @classmethod
+    def get_data(cls, tds, col):
+        r = tds._reader
+        if col.column_type == SYBDATETIMN:
+            size = r.get_byte()
+            if size == 0:
+                return None
+        else:
+            size = col.size
+        if col.column_type == SYBDATETIME or col.column_type == SYBDATETIMN and size == 8:
+            return Datetime.decode(r.readall(Datetime.size))
+
+        elif col.column_type == SYBDATETIME4 or col.column_type == SYBDATETIMN and size == 4:
+            days, time = r.unpack(TDS_DATETIME4)
+            return cls._base_date + timedelta(days=days, minutes=time)
+        else:
+            assert False
+
+    @classmethod
+    def from_python(cls, tds, col, value):
+        col.column_type = SYBDATETIMN
+        col.size = 8
+        Datetime.validate(value)
+        return value
+
+    @staticmethod
+    def get_declaration(tds, col):
+        t = col.column_type
+        if t == SYBDATETIME or t == SYBDATETIMN and col.size == 8:
+            return "DATETIME"
+        elif t == SYBDATETIME4 or t == SYBDATETIMN and col.size == 4:
+            return "SMALLDATETIME"
+
+    @staticmethod
+    def put_info(tds, col):
+        w = tds._writer
+        if col.column_type == SYBDATETIMN:
+            w.put_byte(col.size)
+
+    @classmethod
+    def put_data(cls, tds, col):
+        w = tds._writer
+        value = col.value
+        if col.column_type == SYBDATETIMN:
+            if value is None:
+                w.put_byte(0)
+                return
+            else:
+                w.put_byte(col.size)
+
+        if col.column_type == SYBDATETIME or col.column_type == SYBDATETIMN and col.size == 8:
+            w.write(Datetime.encode(value))
+        elif col.column_type == SYBDATETIME4 or col.column_type == SYBDATETIMN and col.size == 4:
+            raise NotImplementedError
+        else:
+            assert False
 
 #
 # Fetch character data the wire.
@@ -967,3 +1120,31 @@ def tds_get_char_data(tds, wire_size, curcol):
         return curcol.char_codec.decode(r.readall(wire_size))[0]
     else:
         return r.readall(wire_size)
+
+class Datetime:
+    base_date = datetime(1900, 1, 1)
+    min = datetime(1753, 1, 1, 0, 0, 0)
+    max = datetime(9999, 12, 31, 23, 59, 59, 997000)
+    size = 8
+
+    @classmethod
+    def validate(cls, value):
+        if not (cls.min <= value <= cls.max):
+            raise DataError('Date is out of range')
+
+    @classmethod
+    def encode(cls, value):
+        cls.validate(value)
+        if type(value) == date:
+            value = datetime.combine(value, time(0,0,0))
+        days = (value - cls.base_date).days
+        ms = value.microsecond//1000
+        tm = (value.hour * 60 * 60 + value.minute * 60 + value.second)*300 + int(round(ms*3/10.0))
+        return TDS_DATETIME.pack(days, tm)
+
+    @classmethod
+    def decode(cls, buf):
+        days, time = TDS_DATETIME.unpack(buf)
+        ms = int(round(time%300 * 10/3.0))
+        secs = time//300
+        return cls.base_date + timedelta(days=days, seconds=secs, milliseconds=ms)
