@@ -2,7 +2,7 @@
 """DB-SIG compliant module for communicating with MS SQL servers"""
 
 __author__ = 'Mikhail Denisenko <denisenkom@gmail.com>'
-__version__ = '0.6.0'
+__version__ = '0.7.0'
 
 import logging
 import decimal
@@ -149,7 +149,6 @@ class _Connection(object):
             pass
 
     def __del__(self):
-        logger.debug("MSSQLConnection.__del__()")
         if self._conn is not None:
             self._conn.close()
 
@@ -508,7 +507,8 @@ class _Cursor(object):
         """
         Closes the cursor. The cursor is unusable from this point.
         """
-        self._conn._close_cursor(self)
+        if self._conn is not None:
+            self._conn._close_cursor(self)
 
     def execute(self, operation, params=()):
         # Execute the query
@@ -620,7 +620,7 @@ def connect(server='.', database='', user='', password='', timeout=0,
         login_timeout=60, as_dict=False,
         host='', appname=None, port=None, tds_version=TDS74,
         encryption_level=TDS_ENCRYPTION_OFF, autocommit=True,
-        blocksize=4096, use_mars=True):
+        blocksize=4096, use_mars=True, auth=None):
     """
     Constructor for creating a connection to the database. Returns a
     Connection object.
@@ -697,6 +697,7 @@ def connect(server='.', database='', user='', password='', timeout=0,
     login.server_name = server
     login.instance_name = instance
     login.blocksize = blocksize
+    login.auth = auth
     return _Connection(login, as_dict, autocommit)
 
 def Date(year, month, day):
